﻿using System;
using System.Collections.Generic;
using System.Text;

namespace MiscUtil.Linq
{
    /// <summary>
    /// Very simple implementation of IDataProducer.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class DataProducer<T> : IDataProducer<T>
    {
        /// <summary>
        /// Event which is raised when an item of data is produced.
        /// This will not be raised after EndOfData has been raised.
        /// The parameter for the event is the 
        /// </summary>
        /// <seealso cref="MiscUtil.Linq.IDataProducer&lt;T&gt;.DataProduced"/>
        public event Action<T> DataProduced;
        /// <summary>
        /// Event which is raised when the sequence has finished being
        /// produced. This will be raised exactly once, and after all
        /// DataProduced events (if any) have been raised.
        /// </summary>
        /// <seealso cref="MiscUtil.Linq.IDataProducer&lt;T&gt;.EndOfData"/>
        public event Action EndOfData;
        bool endReached = false;

        /// <summary>
        /// Signals a single item of data.
        /// </summary>
        public void Produce(T item)
        {
            if (endReached)
            {
                throw new InvalidOperationException("Cannot produce after end of data");
            }
            if (DataProduced != null)
            {
                DataProduced(item);
            }
        }

        /// <summary>
        /// Signals multiple items of data, one at a time, then ends.
        /// Note that this method only exists to support the params modifier.
        /// In every other way it's equivalent to the ProduceAndEnd(IEnumerable{T}).
        /// </summary>
        public void ProduceAndEnd(params T[] items)
        {
            ProduceAndEnd((IEnumerable<T>)items);
        }

        /// <summary>
        /// Signals multiple items of data, one at a time, then ends.
        /// </summary>
        public void ProduceAndEnd(IEnumerable<T> items)
        {
            foreach (T item in items)
            {
                Produce(item);
            }
            End();
        }

        /// <summary>
        /// Signal the end of data. This can only be called once, and
        /// afterwards the Produce method must not be called.
        /// </summary>
        public void End()
        {
            if (endReached)
            {
                throw new InvalidOperationException("Cannot produce end twice");
            }
            endReached = true;
            if (EndOfData != null)
            {
                EndOfData();
            }
        }
    }
}
